#!/bin/sh

# Copyright 2018, 2019 Paul Hänsch
#
# This is a file format helper, part of CGIlite.
# 
# CGIlite is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# CGIlite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with CGIlite.  If not, see <http://www.gnu.org/licenses/>. 

[ -n "$include_storage" ] && return 0
include_storage="$0"

CR="
"
BR='
'

LOCK(){
  local lock timeout block
  lock="${1}.lock"
  timeout="${2-20}"
  if [ \! -w "${lock%/*}" ] || [ -e "$lock" -a \! -d "$lock" ]; then
    debug "Impossible to get lock: $lock"
    return 1
  fi

  while ! mkdir "$lock" 2>&-; do
    block="$(cat "$lock/pid" || printf 1)"
    if ! { ps -eo pid |grep -qwF "$block"; }; then
      debug "Overriding stale lock: $lock"
      break
    fi
    if [ $timeout -le 0 ]; then
      debug "Timeout while trying to get lock: $lock"
      return 1
    fi
    timeout=$((timeout - 1))
    sleep 1
  done
  printf '%i\n' $$ >"${lock}/pid"
  return 0
}

RELEASE(){
  local lock
  lock="${1}.lock"
  if [ "$(cat "$lock/pid")" = "$$" ]; then
    rm "$lock/pid"
    if ! rmdir "$lock"; then
      debug "Cannot remove tainted lock: $lock"
      printf '%i\n' $$ >"${lock}/pid"
      return 1
    fi
    return 0
  else
    debug "Refusing to release foreign lock: $lock"
    return 1
  fi
}

# STRING='
#   s;\\;\\\\;g; s;\t;\\t;g;
#   s;\n;\\n;g;  s;\r;\\r;g;
#   s;\+;\\+;g;  s; ;+;g;
# '
STRING(){
  local in out=''
  [ $# -gt 0 ] && in="$*" || in="$(cat)"
  while [ "$in" ]; do case $in in
    \\*) out="${out}\\\\"; in="${in#\\}" ;;
    "$BR"*) out="${out}\\n"; in="${in#${BR}}" ;;
    "$CR"*) out="${out}\\r"; in="${in#${CR}}" ;;
    "	"*) out="${out}\\t"; in="${in#	}" ;;
    +*) out="${out}\\+"; in="${in#+}" ;;
    " "*) out="${out}+"; in="${in# }" ;;
    *) out="${out}${in%%[\\${CR}${BR}	+ ]*}"; in="${in#"${in%%[\\${BR}${CR}	+ ]*}"}" ;;
  esac; done
  printf '%s' "$out"
}

UNSTRING='
  :UNSTRING_X
  s;((^|[^\\])(\\\\)*)\\n;\1\n;g;
  s;((^|[^\\])(\\\\)*)\\t;\1\t;g;
  s;((^|[^\\])(\\\\)*)\\r;\1\r;g;
  s;((^|[^\\])(\\\\)*)\+;\1 ;g;
  tUNSTRING_X;
  s;((^|[^\\])(\\\\)*)\\\+;\1+;g;
  s;\\\\;\\;g;
'
UNSTRING(){
  local in out=''
  [ $# -gt 0 ] && in="$*" || in="$(cat)"
  while [ "$in" ]; do case $in in
    \\\\*) out="${out}\\"; in="${in#\\\\}" ;;
    \\n*) out="${out}${BR}"; in="${in#\\n}" ;;
    \\r*) out="${out}${CR}"; in="${in#\\r}" ;;
    \\t*) out="${out}	"; in="${in#\\t}" ;;
    \\+*) out="${out}+"; in="${in#\\+}" ;;
    +*) out="${out} "; in="${in#+}" ;;
    \\*) in="${in#\\}" ;;
    *) out="${out}${in%%[\\+]*}"; in="${in#"${in%%[\\+]*}"}" ;;
  esac; done
  printf '%s' "$out"
}

DBM() {
  local file="$1" cmd="$2"
  local k v key value
  shift 2;

  case "$cmd" in
    check|contains)
      key="$(STRING "$1")"
      while read -r k v; do if [ "$k" = "$key" ]; then
        return 0
      fi; done <"$file" 2>&-
      return 1
      ;;
    get)
      key="$(STRING "$1")"
      while read -r k v; do if [ "$k" = "$key" ]; then
        UNSTRING "$v"
        return 0
      fi; done <"$file" 2>&-
      return 1
      ;;
    set|store)
      key="$(STRING "$1")" value="$(STRING "$2")"
      LOCK "$file" || return 1
      { while read -r k v; do
          [ "$k" = "$key" ] || printf '%s\t%s\n' "$k" "$v"
        done <"$file" 2>&-
        printf '%s\t%s\n' "$key" "$value"
      } >"${file}.$$.tmp"
      mv "${file}.$$.tmp" "${file}"
      RELEASE "$file"
      return 0
      ;;
    add|insert)
      k="$1" key="$(STRING "$1")" value="$(STRING "$2")"
      LOCK "$file" || return 1
      if DBM "$file" check "$k"; then
        RELEASE "$file"
        return 1
      else
        printf '%s\t%s\n' "$key" "$value" >>"${file}"
        RELEASE "$file"
        return 0
      fi
      ;;
    update|replace)
      k="$1" key="$(STRING "$1")" value="$(STRING "$2")"
      LOCK "$file" || return 1
      if ! DBM check "$k"; then
        RELEASE "$file"
        return 1
      fi
      { while read -r k v; do
          [ "$k" = "$key" ] \
          && printf '%s\t%s\n' "$key" "$value" \
          || printf '%s\t%s\n' "$k" "$v"
        done <"$file" 2>&-
      } >"${file}.$$.tmp"
      mv "${file}.$$.tmp" "${file}"
      RELEASE "$file"
      return 0
      ;;
    append)
      key="$(STRING "$1")" value="$(STRING "$2")"
      LOCK "$file" || return 1
      if ! DBM check "$1"; then
        RELEASE "$file"
        return 1
      fi
      { while read -r k v; do
          [ "$k" = "$key" ] \
          && printf '%s\t%s\n' "$key" "$v$value" \
          || printf '%s\t%s\n' "$k" "$v"
        done <"$file" 2>&-
      } >"${file}.$$.tmp"
      mv "${file}.$$.tmp" "${file}"
      RELEASE "$file"
      return 0
      ;;
    delete|remove)
      key="$(STRING "$1")"
      LOCK "$file" || return 1
      { while read -r k v; do
          [ "$k" = "$key" ] || printf '%s\t%s\n' "$k" "$v"
        done <"$file" 2>&-
      } >"${file}.$$.tmp"
      mv "${file}.$$.tmp" "${file}"
      RELEASE "$file"
      return 0
      ;;
  esac
}
